//=============================================================================
// POCKiller
//
// PAWN OBJECT CONTROL KILLER
// This is the mutator and thus the main class of this mod.
//
// It basically intercepts the "IsRelevant()" engine call and acquits all
// actors with "true" back to the engine, until a Pawn trigger is found (by
// default, any PlayerPawn, though you -could- configure any Pawn type).
//
// Then the mutator checks the next number of actors coming through, on the
// assumption that following a Pawn, the server would spawn the objects
// belonging to that pawn soon after it (CheckNumberSpawns is the maximum
// amount of "follow up" actors that are being checked after a pawn was
// encountered).
//
// If any actors are encountered after the triggering pawn which match a
// configured entry, they will be removed - either by telling the game engine
// that they are no longer relevant or by destroying them directly
// (aggressive mode).
//
// Selective destruction (Filter by playername) is a bit more complicated and
// requires a delayed destruction, because the PlayerReplicationInfo may not
// (yet) be available at the time the actor(s) to be removed are encountered.
// After another small delay (which you can also configure:
// SelectiveDestructionDelay), the object is tested for its OWNER, if it is a
// trigger object, if it has a PlayerReplicationInfo, and then a filter by
// name evaluation is done. In case the PlayerReplicationInfo cannot be read,
// the attempt may be repeated after more delays (configured by
// SelectiveDestructionAttempts).
//
//  Author: }TCP{Wolf
//    Date: February 2012
// Version: 1.01
//
// History:
// 1.00 - First version for Sparky to test :)
// 1.01 - changed initialization to happen earlier, and to prevent recursion
//=============================================================================

class POCKiller extends Mutator config(POCKiller);


// CONFIGURATION DEFINITIONS
var struct ObjectsToKill_TYPE
{
	var() config bool bEnabled;         // enable / disable every single entry
	var() config String PackageName;    // the package name
	var() config String ClassName;      // the classname, leave blank if you do not know / unidentifyable
}NONEXISTING_DECLARATION_Dummy;   // The full "loader" is "packagename.classname"

var struct NameFilter_TYPE
{
	var() config bool bEnabled;
	var() config String PlayerName;
}NONEXISTING_DECLARATION2_TYPE;

// CONFIGURATION
var() config bool bDebug;                  // if true... log a lot more crap...
var() config bool bEnabled;                // well... obviously this code won't do much if this is false...
var() config String BaseSearchClass_PAWN;  // Pawn class that will trigger checking for "bad" objects
var() config int CheckNumberSpawns;        // how many
var() config ObjectsToKill_TYPE ObjectsToKill[32]; // The actual list of objects we wish to kill
var() config bool bAggressiveDestruction;  // set to true to actively destroy the object rather than relying on mutator logic
var() config bool bSelective;              // set to true to destroy actors only on certain players (name filter)
var() config float SelectiveDestructionDelay;   // value in seconds to delay selective destruction
var() config int SelectiveDestructionAttempts;  // value to determine how often to try seletive destruction (if PlayerReplicationInfo is not yet available)
var() config NameFilter_TYPE PlayerFilter[64]; // Input all playernames to be filtered by in here


// RUNTIME
var class<Pawn> BaseSearchClass;  // Pawn class that will trigger checking for "bad" objects
var bool bInitDone;               // Originally false, will turn true once initialization is complete
var int SpawnChecksCounter;       // As long as this counter is greater than zero, objects will be looked at
var KillObject killObjectChain;   // Root of the killobject chain...
var int LastPlayerName;           // remember index of last active config...

//=============================================================================
// INITIALIZATION
//=============================================================================


event PreBeginPlay()
{
	init();
	super.PreBeginPlay();
}

function init()
{
	local bool bSaveConfig;
	local int i;
	local KillObject ko;
	local float f;

	if (bInitDone) return;
	bInitDone = True;

	l("START INITIALIZATION - instance = "$self);

	if (!bEnabled)
	{
		l("Aborting initialization, POCKiller is DISABLED.");
		return;
	}

	bSaveConfig = False;	// This will be set to true by failed sanity checks or failed events during init

	// basic init
	SpawnChecksCounter = 0;

	// Try to load the base class that triggers the checks...
	BaseSearchClass = class<Pawn>(DynamicLoadObject(BaseSearchClass_PAWN, class'Class'));
	if (BaseSearchClass == none)	// hmm that failed...
	{
		l("The specified base pawn class of "$BaseSearchClass_PAWN$" could not be loaded!");
		BaseSearchClass_PAWN = "Engine.PlayerPawn";
		l("Using internal default value...");
		BaseSearchClass = class<Pawn>(DynamicLoadObject(BaseSearchClass_PAWN, class'Class'));
		bSaveConfig = True;
	}
	l("Base Pawn Class = "$BaseSearchClass);

	// Sanity checks
	if (CheckNumberSpawns < 3)
	{
		l("CheckNumberSpawns has a minimum value of 3... correcting...");
		CheckNumberSpawns = 3;
		bSaveConfig = True;
	}

	if (bSelective)
	{
		if (SelectiveDestructionDelay <= 0)
		{
			l("SelectiveDestructionDelay zero or lower... correcting to internal default of 1 second...");
			SelectiveDestructionDelay = 1.0;
			bSaveConfig = True;
		}

		if (SelectiveDestructionAttempts < 1)
		{
			l("SelectiveDestructionAttempts min value is ONE...");
			SelectiveDestructionAttempts = 1;
			bSaveConfig = True;
		}

		// ... and some sanity...
		f = SelectiveDestructionAttempts; // float(SelectiveDestructionAttempts)
		f *= SelectiveDestructionDelay;
		//if ( (float(SelectiveDestructionAttempts) * SelectiveDestructionDelay) > 20)
		if (f > 20)
		{
			l("NOTICE: Total configured delay can reach "$f$" seconds!");
			l("NOTICE: You _might_ want to use bDebug and determine if you truly need such large delays...");
		}

		// get index of last active player name filter...
		LastPlayerName = -1;
		for(i=ArrayCount(PlayerFilter)-1; (i>=0) && (LastPlayerName == -1) ;i--)
		{
			if (PlayerFilter[i].bEnabled)
				LastPlayerName = i;
		}

		if (LastPlayerName == -1)
		{
			l("Selective mode was chosen but all player name filters are OFF, disabling selective mode.");
			bSelective = False;
			bSaveConfig = True;
		}
	}

	// setting up kill object chain
	for(i=0; i<ArrayCount(ObjectsToKill); i++) if (ObjectsToKill[i].bEnabled)
	{
		ko = spawn(Class'KillObject');
		// give it required information...
		ko.mut = self;
		ko.PackageName = ObjectsToKill[i].PackageName;

		if (ObjectsToKill[i].ClassName != "")
		{
			ko.actualClass = Class<Actor>(DynamicLoadObject(ObjectsToKill[i].PackageName$"."$ObjectsToKill[i].ClassName, class'Class'));
			if (ko.actualClass == none)
			{
				bSaveConfig = True;
				ObjectsToKill[i].bEnabled = False;
				l("Objects to kill entry "$i$" appears to be bad, class loading failed.");
				l("The entry ["$ObjectsToKill[i].PackageName$"."$ObjectsToKill[i].ClassName$"] will be disabled...");
			}
			else // class loading was successful
				AddKillObject(ko);
		}
		else if (ObjectsToKill[i].PackageName != "")
			AddKillObject(ko);	// add it as "unidentifyabl" class
	} // end for

	// check kill object chain setup...
	if (killObjectChain == none)
	{
		bEnabled = False;
		l("No objects to kill configured... do you need this mod...?");
		l("POCKiller will DISABLE itself until you set it up right!");
	}
	else
	{
		l("The following classes are configured:");
		killObjectChain.ListKillObjects(1);
	}

	// init is basically done... was anything corrected? If yes, save it...
	if (bSaveConfig)
	{
		l("One or more configuration parameters were changed during initialization.");
		l("Saving corrected configuration.");
		SaveConfig();
	}

	l("END INITIALIZATION");
}


function AddKillObject(KillObject newOne)
{
	if (killObjectChain == none) killObjectChain = newOne;
	else killObjectChain.AddKillObject(newOne);
}



//=============================================================================
// MUTATOR ENGINE
//=============================================================================


function bool IsRelevant(Actor Other, out byte bSuperRelevant)
{
	local bool bRelevant;
	local POCDelayedDestroy delayer;

	if (!bInitDone) init(); // lazy init...

	if ( (!bEnabled) ||
				(Other.class == class'POCDelayedDestroy') ||	// own object
				(Other.class == class'KillObject') ||			// own object
				(killObjectChain == none)		// something is nooot right...
		)
		return super.IsRelevant(Other, bSuperRelevant); // nothing todo if disabled...


	if (SpawnChecksCounter >0)
	{
		if (bDebug) l("IsRelevant ["$SpawnChecksCounter$"] "$Other.class);
		SpawnChecksCounter--;
		bRelevant = killObjectChain.IsRelevant(Other);
		if (!bRelevant)
		{
			if (bSelective)
			{
				class'POCDelayedDestroy'.default.theTimer = SelectiveDestructionDelay;
				class'POCDelayedDestroy'.default.mut = self;
				delayer = spawn(class'POCDelayedDestroy',self);
				delayer.suspiciousObject = Other;
				if (bDebug) l("Registering for delayed investigation: "$Other);
				return Super.IsRelevant(Other, bSuperRelevant);
			}

			if (bAggressiveDestruction)
			{
				l("Aggressively destroying: "$Other);
				Other.destroy();
			}
			else
				l("Telling game mode object is irrelevant: "$Other);

			return false;
		}
	}

	if (isTriggerObject(Other))	// the trigger class/subclass is met...
		SpawnChecksCounter += CheckNumberSpawns;	// so let us investigate the next X number of objects...

	// nothing here...
	return super.IsRelevant(Other, bSuperRelevant);
}


//=============================================================================
// AID METHODS
//=============================================================================

// Externalized method from IsRelevant() so available to outside
function bool isTriggerObject(Actor Other)
{
	// sanity
	if (Other == none)
		return False;

	if (ClassIsChildOf(Other.class,BaseSearchClass))	// the trigger class/subclass is met...
	{
		if (bDebug) l(Other.class$" [TRIGGER] - has ReplicationInfo: "$(Pawn(Other).PlayerReplicationInfo != none));
		return True;
	}

	return False;
}


// This is just here for convenience
function bool playerNameMatch(String queryName)
{
	local int i;

	for (i=0; i<=LastPlayerName; i++)
	{
		if (PlayerFilter[i].bEnabled && (PlayerFilter[i].PlayerName ~= queryName) )
		{
			if (bDebug) l("Player name of "$queryName$" matches entry "$i);
			return True;
		}
	}

	return False;
}


//=============================================================================
// STUPID METHODS
//=============================================================================

// Simply log out stuff with a prefix...
function l (String what)
{
	log(what,'POCKiller');
}

defaultproperties
{
	CheckNumberSpawns=10
	bDebug=False
	bEnabled=True
	BaseSearchClass_PAWN="Engine.PlayerPawn"
	bInitDone=False
	bAggressiveDestruction=False
	bSelective=False
	SelectiveDestructionDelay=1.000
	SelectiveDestructionAttempts=2

	ObjectsToKill(0)=(bEnabled=True,PackageName="UTeamFix",ClassName="")
	ObjectsToKill(1)=(bEnabled=False,PackageName="SomeOtherPackage",ClassName="AKnownClass")

	PlayerFilter(0)=(bEnabled=False,PlayerName="AnInnocentPlayerNameExample")

	bHiddenEd=True
	DrawType=DT_None
	Style=STY_None
	Texture=None
	Fatness=0
}
